#include <fstream>
#include <GL/glut.h>
#include <assert.h>
#include "Vector3.h"
#include "BoundingSphere.h"
#include "Mesh.h"

using namespace std;

Mesh::Mesh(const char* filename)
{
   //initialization
   max_x = max_y = max_z = -999999;
   min_x = min_y = min_z = 99999;
   center.x = center.y = center.z = 0;

   //printf("Reading coordinates from %s\n", filename);
  
   ifstream in_f(filename);
   if (!in_f)
      printf("Could not open file %s\n", filename);
   else
      readStream(in_f);

   //once the file is parsed find out the maximum extent to center and scale mesh
   max_extent = max_x - min_x;
   if (max_y - min_y > max_extent) max_extent = max_y - min_y;
    
   center.x /= vertices.size();
   center.y /= vertices.size();
   center.z /= vertices.size();

   setupList();
}

BoundingVolume* Mesh::getBoundingVolume() const
{ return new BoundingSphere(0.6f); }

//process the input stream from the file
void Mesh::readStream(istream& is)
{
   char str[256];

   while (is)
   {
      is >> ws;
      is.get(str,sizeof(str));
      if (!is) break;
      is.ignore(9999,'\n');
      readLine(str);
   }
}

//process each line of input save vertices and faces appropriately
void Mesh::readLine(char* str)
{
   int indx = 0, vi;
   float x, y, z;
   float r, g, b;
   int mat;

   if (str[0] == '#') return;

   if (str[0] == 'V' && !strncmp(str, "Vertex ", 7))
   {
      int count = sscanf(str,"Vertex %d %g %g %g",&vi,&x,&y,&z);
	  assert(count == 4);
      static Vector3f v;
      v.set(x, y, z);

      vertices.push_back(v);
      //cout << "made a vertex " << v.x << " " << v.y << " " << v.z << endl;

      //house keeping to display in center of the scene
      center.x += v.x;
      center.y += v.y;
      center.z += v.z;

      if (v.x > max_x) max_x = v.x; if (v.x < min_x) min_x = v.x;
      if (v.y > max_y) max_y = v.y; if (v.y < min_y) min_y = v.y;
      if (v.z > max_z) max_z = v.z; if (v.z < min_z) min_z = v.z;
   }
   else if (str[0]=='F' && !strncmp(str,"Face ",5)) 
   {
      Triangle t;
      char* s = str + 4;
      int fi = -1;
      for (int t_i = 0;;t_i++)
     {
         while (*s && isspace(*s)) s++;
         
         //if we reach the end of the line break out of the loop
         if (!*s) break;
         
         //save the position of the current character
         char* beg=s;
         
         //advance to next space
         while (*s && isdigit(*s)) s++;

         //covert the character to an integer
         int j = atoi(beg);
      
         //the first number we encounter will be the face index, don't store it      
         if (fi < 0) { fi = j; continue; }

         //otherwise process the digit we've grabbed in j as a vertex index
         //the first number will be the face id the following are vertex ids
         if (t_i == 1)
            t.indices.x = j;
         else if (t_i == 2)
            t.indices.y = j;
         else if (t_i == 3)
             t.indices.z = j;

			//if there is more data to process break out
			if (*s =='{') break;
		}

		//normal vectors
		Vector3f vector1(vertices[t.indices.z-1] - vertices[t.indices.x-1]);
		Vector3f vector2(vertices[t.indices.y-1] - vertices[t.indices.x-1]);
		crossProduct(vector2, vector1, &t.normal);
		t.normal.normalize();
          
		//possibly process colors if the mesh has colors
		if (*s && *s =='{')
		{
			char *s1 = s+1;

			if (!strncmp(s1,"rgb=",4))
			{
				//grab the values of the string
				int count = sscanf(s1,"rgb=(%g %g %g) matid=%d",&r,&g,&b,&mat);
				assert(count == 4);
				t.color.set(r, g, b);
			}
			else //set default color
				t.color.set(0.5f, 0.5f, 0.5f);
		}
		else //set default color
			t.color.set(0.5f, 0.5f, 0.5f);

		triangles.push_back(t);
	}
}

void Mesh::setupVertices()
{
   //leave these transformations in as they center and scale each mesh correctly
   //scale object to window
   glScalef(1.0/(float)max_extent, 1.0/(float)max_extent, 1.0/(float)max_extent);
   //translate the object to the orgin
   glTranslatef(-(center.x), -(center.y), -(center.z));

   //draw the mesh
   for (unsigned int j = 0; j < triangles.size(); j++)
      drawTriangle(triangles[j]);
}

//drawing routine to draw triangles as wireframe
void Mesh::drawTriangle(Triangle t) const
{
   glBegin(GL_TRIANGLES);
   glColor4f(t.color.x, t.color.y, t.color.z, 1.0f);

   //note that the vertices are indexed starting at 0, but the triangles
   //index them starting from 1, so we must offset by -1!!!
   glNormal3f(t.normal.x, t.normal.y, t.normal.z);
   glVertex3f(vertices[t.indices.x - 1].x, 
              vertices[t.indices.x - 1].y,
              vertices[t.indices.x - 1].z);

   glVertex3f(vertices[t.indices.y - 1].x, 
              vertices[t.indices.y - 1].y,
              vertices[t.indices.y - 1].z);
 
   glVertex3f(vertices[t.indices.z - 1].x, 
              vertices[t.indices.z - 1].y,
              vertices[t.indices.z - 1].z);

   glEnd();
}
