#include <iostream>
#include <GL/glut.h>
#include "Geometry.h"
#include "BoundingVolume.h"
#include "Player.h"
#include "math_util.h"
#include "Projectile.h"

#define USE_SHADERS
//#define MULTI

#ifdef MULTI
#include "Message.h"
#include "ClientSocket.h"
extern ClientSocket connection;
#endif

#ifdef USE_SHADERS
#include "PlayerShield.h"
#endif

const float Player::MaxVelocity = 10.0f;
const float Player::MaxEnergy = 100.0f;
const float Player::MaxShield = 100.0f;
const float Player::CollisionDamage = 5.0f;

Player::Player(Geometry *g, const Vector3f& startPos, const Vector3f& startDir) 
	: Object3d(g, startPos)
{
	velocity = 0.0f;
	actualVelocity = 0.0f;
	shield = MaxShield;
	energy = MaxEnergy;
	shieldRegen = 1.0f;
	energyRegen = 10.0f;
	acceleration = 0.1f;
	alive = true;

	currentTime = 0.0f;
	collForce.set(0.0f, 0.0f, 0.0f);
	leftWeapon = rightWeapon = 0;

	rotation.set(Vector3f(1.0f, 0.0f, 0.0f), startDir);

	setAxisVectors(startDir, Vector3f(0.0f, 1.0f, 0.0f));

#ifdef USE_SHADERS
	setBoundVolume(new PlayerShield(&position, 0.6f, "shield.cg"));
#endif
}

Player::~Player()
{
	if (leftWeapon)
	{
		delete leftWeapon; 
		leftWeapon = 0;
	}

	if (rightWeapon)
	{
		delete rightWeapon;
		rightWeapon = 0;
	}
}

void Player::setAxisVectors(const Vector3f& dir, const Vector3f& up)
{
	rotAxis.z = dir;
	rotAxis.z.normalize();
	rotAxis.y = up;
	rotAxis.y.normalize();
	crossProduct(rotAxis.z, rotAxis.y, &rotAxis.x);
	rotAxis.x.normalize();
}

void Player::updateFromServer(char* buffer)
{
	#ifdef MULTI
	Vector3f dir, up;
	Message::ParseAll(buffer, &position, &dir, &up, &velocity, &rotation);
	setAxisVectors(dir, up);
	#endif
}

void Player::die()
{
	alive = false;
}

void Player::update(float time)
{
	if (!alive)	return;

	currentTime += time;
	energy += (energyRegen * time);
	if (energy > MaxEnergy) energy = MaxEnergy;
	shield += (shieldRegen * time);
	if (shield > MaxShield) shield = MaxShield;

	// Constant acceleration
	if (!equals(velocity, actualVelocity))
	{
		if (actualVelocity < velocity)
			actualVelocity = min(velocity, actualVelocity+acceleration);
		else
			actualVelocity = max(velocity, actualVelocity-acceleration);
	}

	Vector3f oldPosition = position;
	position += rotAxis.z * actualVelocity * time + collForce * time;
	collForce.set(0.0f, 0.0f, 0.0f);

	// Collision detection
	ObjectIter objIter = Object3d::objects.begin();
	ObjectIter objEndIter = Object3d::objects.end();

	for (; objIter != objEndIter; objIter++)
	{
		if (*objIter == this) continue;

		const BoundingVolume *otherVol = (*objIter)->getBoundVolume();

		// If Collision
		if (bVolume->intersects(otherVol))
		{
			shield -= CollisionDamage;

			if (shield < 0)
				die();

			#ifdef USE_SHADERS
				(dynamic_cast<PlayerShield*>(bVolume))->setImpact(otherVol->getPosition() - position, currentTime);
			#endif

			// If we would move away from the obstacle, just move normally
			if (position.distance(otherVol->getPosition()) > oldPosition.distance(otherVol->getPosition()))
				break;

			position = oldPosition;

			collForce.set(position - otherVol->getPosition());
            collForce.normalize();
			collForce.scale(velocity);

			actualVelocity = -actualVelocity;
			//newPosition = position + rotAxis.z * actualVelocity * time + collForce * time;

			return;
		}
	}
}

void Player::yaw(float amount)
{
	rotAxis.x.rotate(amount, rotAxis.y);
	rotAxis.x.normalize();
	rotAxis.z.rotate(amount, rotAxis.y);
	rotAxis.z.normalize();

	rotation *= Quaternion(rotAxis.y, amount);
	leftWeapon->rotate(amount, rotAxis.y);
	rightWeapon->rotate(amount, rotAxis.y);
}

void Player::pitch(float amount)
{
	rotAxis.y.rotate(amount, rotAxis.x);
	rotAxis.y.normalize();

	rotAxis.z.rotate(amount, rotAxis.x);
	rotAxis.z.normalize();

	rotation *= Quaternion(rotAxis.x, amount);
	leftWeapon->rotate(amount, rotAxis.x);
	rightWeapon->rotate(amount, rotAxis.x);
}

void Player::roll(float dir)
{
	rotAxis.x.rotate(dir, rotAxis.z);
	rotAxis.x.normalize();

	rotAxis.y.rotate(dir, rotAxis.z);
	rotAxis.y.normalize();

	rotation *= Quaternion(rotAxis.z, dir);
	leftWeapon->rotate(dir, rotAxis.z);
	rightWeapon->rotate(dir, rotAxis.z);
}

void Player::draw() const
{
	if (alive)
		Object3d::draw();
}

void Player::drawBoundingVolume()
{
#ifdef USE_SHADERS
	if (visible)
		(dynamic_cast<PlayerShield*>(bVolume))->drawShield(currentTime);
#endif
}

void Player::hit(const Projectile* projectile, const Vector3f& location)
{ 
	shield -= projectile->getPower();

#ifdef USE_SHADERS
	(dynamic_cast<PlayerShield*>(bVolume))->setImpact(location-position, currentTime);
#endif
}

void Player::fireFirst()
{
	if (energy > leftWeapon->energyDrain)
	{
		energy -= leftWeapon->energyDrain;
		leftWeapon->fire();		
	}
}

void Player::fireSecond()
{
	if (energy > rightWeapon->energyDrain)
	{
		energy -= rightWeapon->energyDrain;
		rightWeapon->fire();		
	}
}

void Player::fireOther(int type)
{
	setPrimaryWeapon(type);
	leftWeapon->fireOther();
	delete leftWeapon;
}

void Player::setPrimaryWeapon(int type) {
   Vector3f left_offset, right_offset;
   Spacecraft *craft;

   craft = dynamic_cast<Spacecraft*>(geometry);
   left_offset = craft->getLeftWeaponOffset();
   right_offset = craft->getRightWeaponOffset();
   //if (leftWeapon) delete leftWeapon;
   leftWeapon = new Weapon(this, left_offset, right_offset, type);
}

void Player::setSecondaryWeapon(int type) {
   Vector3f left_offset, right_offset;
   Spacecraft *craft;

   craft = dynamic_cast<Spacecraft*>(geometry);
   left_offset = craft->getLeftWeaponOffset();
   right_offset = craft->getRightWeaponOffset();
   //if (rightWeapon) delete rightWeapon;
   rightWeapon = new Weapon(this, left_offset, right_offset, type);
}
