#include <cstdio>
#include <iostream>
#include "Shader.h"

CGcontext Shader::cgContext;
CGprofile Shader::cgVertexProfile;

bool Shader::Init()
{
	// Create A New Context For Our Cg Program(s)
	cgContext = cgCreateContext();

	// Validate Our Context Generation Was Successful
	if (cgContext == 0)
	{
		fprintf(stderr, "Failed To Create Cg Context\n");
		return false;
	}

	// Get The Latest GL Vertex Profile
	cgVertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);

	// Validate Our Profile Determination Was Successful
	if (cgVertexProfile == CG_PROFILE_UNKNOWN)
	{
		fprintf(stderr, "Invalid profile type\n");
		return false;
	}

	return true;
}

Shader::Shader(const char *filename)
{
	// Set The Current Profile	
	cgGLSetOptimalOptions(cgVertexProfile);

	// Load And Compile The Vertex Shader From File
	cgProgram = cgCreateProgramFromFile(cgContext, CG_SOURCE, filename, cgVertexProfile, "main", 0);

	// Validate Success
	if (cgProgram == 0)
	{
		// We Need To Determine What Went Wrong
		CGerror Error = cgGetError();

		if (Error == CG_COMPILER_ERROR)
			fprintf(stderr, "Shader Compile Error:\n %s \n", cgGetLastListing(cgContext));
		else
			fprintf(stderr, "Shader(): %s \n", cgGetErrorString(Error));
		enabled = false;
		return;
	}

	// Load The Program
	cgGLLoadProgram(cgProgram);

	modelViewMatrix	= cgGetNamedParameter(cgProgram, "ModelViewProj");
	modelViewIT = cgGetNamedParameter(cgProgram, "ModelViewIT");

	RecurseParameters(cgGetFirstParameter(cgProgram, CG_PROGRAM));

	//fprintf(stderr,"Shader has %d params\n", params.size());
	enabled = true;
}

void Shader::RecurseParameters(CGparameter parameter)
{
	if (parameter == 0)
		return;
	do
	{
		switch(cgGetParameterType(parameter)) 
		{
			case CG_STRUCT:
			{
				RecurseParameters(cgGetFirstStructParameter(parameter));
				break;
			}
			case CG_ARRAY:
			{
				int arraySize = cgGetArraySize(parameter, 0);
				for (int i = 0; i < arraySize; ++i)
					RecurseParameters(cgGetArrayParameter(parameter, i));
				break;
			}
			default:
			{
				params.push_back(parameter);
				break;
			}
		}
	} while ((parameter = cgGetNextParameter(parameter))!= 0);
}

void Shader::setFloatParam(int paramIndex, float value)
{
	//std::cout << "Setting " << cgGetParameterName(params[paramIndex]) << " to " << value << std::endl;
	cgGLSetParameter1f(params[paramIndex], value);
}

void Shader::setVector3fParam(int paramIndex, const Vector3f& value)
{
	//std::cout << "Setting " << cgGetParameterName(params[paramIndex]) << " to " << value << std::endl;
	cgGLSetParameter3f(params[paramIndex], value.x, value.y, value.z);
}

void Shader::setVector4fParam(int paramIndex, const Vector3f& vec, float w)
{
	//std::cout << "Setting " << cgGetParameterName(params[paramIndex]) << " to " 
	//	<< vec << " " << w << std::endl;
	cgGLSetParameter4f(params[paramIndex], vec.x, vec.y, vec.z, w);
}

void Shader::setVector4fParam(int paramIndex, float x, float y, float z, float w)
{
	//std::cout << "Setting " << cgGetParameterName(params[paramIndex]) << " to " 
	//	<< x << " " << y << " " << z << " " << w << std::endl;
	cgGLSetParameter4f(params[paramIndex], x, y, z, w);
}

void Shader::setNamedArray4fParam(const char* paramName, int numElements, float *array)
{
	//std::cout << "Setting " << cgGetParameterName(params[paramIndex]) << " to " 
	//	<< x << " " << y << " " << z << " " << w << std::endl;
	CGparameter param = cgGetNamedParameter(cgProgram, paramName);
	cgGLSetParameterArray4f(param, 0, numElements, array);
}

void Shader::enable() const
{
	// Set The Modelview Matrix Of Our Shader To Our OpenGL Modelview Matrix
	cgGLSetStateMatrixParameter(modelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
	cgGLSetStateMatrixParameter(modelViewIT, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_INVERSE_TRANSPOSE);

	// Enable Our Vertex Shader Profile
	cgGLEnableProfile(cgVertexProfile);

	// Bind Our Vertex Program To The Current State
	cgGLBindProgram(cgProgram);

	// Set the shader parameters
	//cgGLSetParameter4f(color, 1.0f, 0.1f, 0.1f, 1.0f);
	//cgGLSetParameter3f(color, 1.0f, 0.1f, 0.1f);

	//draw geometry here
}

void Shader::disable() const
{
	cgGLDisableProfile(cgVertexProfile);
}

void Shader::Deinitialize()
{
	// Destroy Our Cg Context And All Programs Contained Within It
	cgDestroyContext(cgContext);
}