#ifndef VECTOR3_H
#define VECTOR3_H

#include <math.h>
#include <string>
#include <assert.h>

struct Quaternion;

template <class Type>
struct Vector3
{
	Type x, y, z;

	// Default constructor, does not initialize for efficiency reasons
	Vector3<Type>() { }

	Vector3<Type>(const Type& x, const Type& y, const Type& z)
	{ set(x, y, z); }

	void set(const Vector3<Type>& vec) { x = vec.x; y = vec.y; z = vec.z; }
	void set(const Type& x, const Type& y, const Type& z);

	Type length() const
	{ return (Type) sqrt(x*x + y*y + z*z); }

	void scale(const Type & scalar)
	{ x *= scalar; y *= scalar; z *= scalar; }

	void normalize()
	{ assert(length() > 0); scale(Type(1.0)/length()); }

	// Same as multiplying the vector by -1
	void invert() 
	{ x = -x; y = -y; z = -z; }

	// Rotates this vector along the axis
	void rotate(const Type& angle, const Vector3<Type>& axis);
	void rotate(const Quaternion& quaternion);

	// Increments every value in this vector by the corresponding value 
	// in the argument vector
	void increment(const Vector3<Type> & vec);
	void operator+=(const Vector3<Type> & vec) 
	{ increment(vec); }

	// Decrements every value in this vector by the corresponding value 
	// in the argument vector
	void decrement(const Vector3<Type> & vec);
	void operator-=(const Vector3<Type> & vec) 
	{ decrement(vec); }
	
	// Returns the distance from this vector to the argument vector
	Type distance(const Vector3<Type>& vec) const;

	// Returns the squared distance from this vector to the argument vector
	Type sqDistance(const Vector3<Type>& vec) const;

	Vector3<Type> operator-() const
	{ Vector3<Type> vec(*this); vec.invert(); return vec; }
};

template <class Type>
void Vector3<Type>::set(const Type& x, const Type& y, const Type& z)
{
	this->x = x;
	this->y = y;
	this->z = z;
}

template <class Type>
void Vector3<Type>::rotate(const Type& angle, const Vector3<Type>& axis)
{
	Type cosAngle = cos(angle);
	Type sinAngle = sin(angle);

	Vector3<Type> newVector;
	newVector.x  = (cosAngle + (Type(1) - cosAngle) * axis.x * axis.x)	* x;
	newVector.x += ((Type(1) - cosAngle) * axis.x * axis.y - axis.z * sinAngle) * y;
	newVector.x += ((Type(1) - cosAngle) * axis.x * axis.z + axis.y * sinAngle) * z;
	newVector.y  = ((Type(1) - cosAngle) * axis.x * axis.y + axis.z * sinAngle) * x;
	newVector.y += (cosAngle + (Type(1) - cosAngle) * axis.y * axis.y)			* y;
	newVector.y += ((Type(1) - cosAngle) * axis.y * axis.z - axis.x * sinAngle) * z;
	newVector.z  = ((Type(1) - cosAngle) * axis.x * axis.z - axis.y * sinAngle) * x;
	newVector.z += ((Type(1) - cosAngle) * axis.y * axis.z + axis.x * sinAngle) * y;
	newVector.z += (cosAngle + (Type(1) - cosAngle) * axis.z * axis.z)			* z;

	set(newVector.x, newVector.y, newVector.z);
}
Quaternion operator*(const Quaternion& A, const Quaternion& B);

template <class Type>
void Vector3<Type>::rotate(const Quaternion& quaternion)
{
	Quaternion conjugate = quaternion.conjugate();
	Quaternion vector(*this);
	Quaternion newVector = quaternion * vector * conjugate;

	assert (equals(newVector.w, 0.0f));

	x = newVector.x;
	y = newVector.y;
	z = newVector.z;
}

template <class Type>
void Vector3<Type>::increment(const Vector3<Type> & vec)
{ x += vec.x; y += vec.y; z += vec.z; }

template <class Type>
void Vector3<Type>::decrement(const Vector3<Type> & vec)
{ x -= vec.x; y -= vec.y; z -= vec.z; }

template <class Type>
Type Vector3<Type>::distance(const Vector3<Type>& vec) const
{ 
	return sqrt(sqDistance(vec)); 
}

template <class Type>
Type Vector3<Type>::sqDistance(const Vector3<Type>& vec) const
{ 
	return ((x - vec.x)*(x - vec.x) +
			(y - vec.y)*(y - vec.y) +
			(z - vec.z)*(z - vec.z)); 
}

template <class Type>
void crossProduct(const Vector3<Type>& first, const Vector3<Type>& second, Vector3<Type>* result)
{
	result->x = first.y * second.z - first.z * second.y;
	result->y = first.z * second.x - first.x * second.z;
	result->z = first.x * second.y - first.y * second.x;
}

template <class Type>
Type dotProduct(const Vector3<Type>& first, const Vector3<Type>& second)
{ return first.x * second.x + first.y * second.y + first.z * second.z; }

template <class Type>
Vector3<Type> operator+(const Vector3<Type> & first, const Vector3<Type> & second)
{ return Vector3<Type>(first.x + second.x, first.y + second.y, first.z + second.z); }

template <class Type>
Vector3<Type> operator-(const Vector3<Type> & first, const Vector3<Type> & second)
{ return Vector3<Type>(first.x - second.x, first.y - second.y, first.z - second.z); }
   
template <class Type>
Vector3<Type> operator*(const Vector3<Type> & vec, const Type & scalar)
{ return Vector3<Type>(vec.x * scalar, vec.y * scalar, vec.z * scalar); }

// Can be used with cout to print out vectors as (x, y, z)
// e.g. cout << ZeroVector will print out (0.0, 0.0, 0.0)
template <class Type>
std::ostream& operator<<(std::ostream& out, const Vector3<Type>& vec)
{ out << "(" << vec.x << ", " << vec.y << ", " << vec.z << ")"; return out; }

// A vector of three ints
typedef Vector3<int> Vector3i;

// A vector of three floats
typedef Vector3<float> Vector3f;

#endif